/**
 * @file llgltfbufferutil.h
 * @brief LLGLTF Implementation
 *
 * $LicenseInfo:firstyear=2024&license=viewergpl$
 *
 * Copyright (c) 2024, Linden Research, Inc.
 *
 * Second Life Viewer Source Code
 * The source code in this file ("Source Code") is provided by Linden Lab
 * to you under the terms of the GNU General Public License, version 2.0
 * ("GPL"), unless you have obtained a separate licensing agreement
 * ("Other License"), formally executed by you and Linden Lab.  Terms of
 * the GPL can be found in doc/GPL-license.txt in this distribution, or
 * online at http://secondlifegrid.net/programs/open_source/licensing/gplv2
 *
 * There are special exceptions to the terms and conditions of the GPL as
 * it is applied to this Source Code. View the full text of the exception
 * in the file doc/FLOSS-exception.txt in this software distribution, or
 * online at
 * http://secondlifegrid.net/programs/open_source/licensing/flossexception
 *
 * By copying, modifying or distributing this software, you acknowledge
 * that you have read and understood your obligations described above,
 * and agree to abide by those obligations.
 *
 * ALL LINDEN LAB SOURCE CODE IS PROVIDED "AS IS." LINDEN LAB MAKES NO
 * WARRANTIES, EXPRESS, IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY,
 * COMPLETENESS OR PERFORMANCE.
 * $/LicenseInfo$
 */

// Inline template implementations for copying data out of glTF buffers.
// DO NOT include from header files to avoid the need to rebuild the whole
// project whenever we add support for more types.

#ifndef LL_LLGLTFBUFFERUTIL_H
#define LL_LLGLTFBUFFERUTIL_H

#include <system_error>		// For std::error_code

#include "boost/json.hpp"

#include "llcolor4u.h"
#include "llgltfasset.h"
#include "llgltfglm.h"
#include "llgltfprimitive.h"
#include "llmath.h"
#include "llstrider.h"
#include "llvector2.h"
#include "llvector3.h"

// Suppress unused function warning: clang complains here but these
// specializations are definitely used.
#if LL_CLANG
# pragma clang diagnostic push
# pragma clang diagnostic ignored "-Wunused-function"
#endif

namespace LLGLTF
{
	class Material;
	class Primitive;

	LL_NO_INLINE void not_implemented(const char* func_sig);

	// Copy one Scalar from src to dst
	template<class S, class T>
	LL_INLINE void copyScalar(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one vec2 from src to dst
	template<class S, class T>
	LL_INLINE void copyVec2(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one vec3 from src to dst
	template<class S, class T>
	LL_INLINE void copyVec3(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one vec4 from src to dst
	template<class S, class T>
	LL_INLINE void copyVec4(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one mat2 from src to dst
	template<class S, class T>
	LL_INLINE void copyMat2(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one mat3 from src to dst
	template<class S, class T>
	LL_INLINE void copyMat3(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	// Copy one mat4 from src to dst
	template<class S, class T>
	LL_INLINE void copyMat4(S* src, T& dst)
	{
		not_implemented(LL_FUNC);
	}

	//=========================================================================
	// Concrete implementations for different types of source and destination
	//=========================================================================

	template<>
	LL_INLINE void copyScalar<F32, F32>(F32* src, F32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U32, U32>(U32* src, U32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S32, S32>(S32* src, S32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U32, U16>(U32* src, U16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S32, S16>(S32* src, S16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U16, U16>(U16* src, U16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S16, S16>(S16* src, S16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U16, U32>(U16* src, U32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S16, S32>(S16* src, S32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U8, U16>(U8* src, U16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S8, S16>(S8* src, S16& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U8, U32>(U8* src, U32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S8, S32>(S8* src, S32& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U8, U64>(U8* src, U64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S8, S64>(S8* src, S64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U16, U64>(U16* src, U64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S16, S64>(S16* src, S64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<U32, U64>(U32* src, U64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyScalar<S32, S64>(S32* src, S64& dst)
	{
		dst = *src;
	}

	template<>
	LL_INLINE void copyVec2<F32, LLVector2>(F32* src, LLVector2& dst)
	{
		dst.set(src[0], src[1]);
	}

	template<>
	LL_INLINE void copyVec3<F32, glm::vec3>(F32* src, glm::vec3& dst)
	{
		dst = glm::vec3(src[0], src[1], src[2]);
	}

	template<>
	LL_INLINE void copyVec3<F32, LLVector4a>(F32* src, LLVector4a& dst)
	{
		dst.load3(src);
	}

	template<>
	LL_INLINE void copyVec3<U16, LLColor4U>(U16* src, LLColor4U& dst)
	{
		dst.set(src[0], src[1], src[2], 255);
	}

	template<>
	LL_INLINE void copyVec4<U8, LLColor4U>(U8* src, LLColor4U& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<U16, LLColor4U>(U16* src, LLColor4U& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<F32, LLColor4U>(F32* src, LLColor4U& dst)
	{
		dst.set(src[0]*255, src[1]*255, src[2]*255, src[3]*255);
	}

	template<>
	LL_INLINE void copyVec4<F32, LLVector4a>(F32* src, LLVector4a& dst)
	{
		dst.loadua(src);
	}

	template<>
	LL_INLINE void copyVec4<U64, LLVector4a>(U64* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<S64, LLVector4a>(S64* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<U32, LLVector4a>(U32* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<S32, LLVector4a>(S32* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<U16, LLVector4a>(U16* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<S16, LLVector4a>(S16* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<U8, LLVector4a>(U8* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<S8, LLVector4a>(S8* src, LLVector4a& dst)
	{
		dst.set(src[0], src[1], src[2], src[3]);
	}

	template<>
	LL_INLINE void copyVec4<F32, glm::quat>(F32* src, glm::quat& dst)
	{
		dst.x = src[0];
		dst.y = src[1];
		dst.z = src[2];
		dst.w = src[3];
	}

	// Used by glTF-Sample-Models-master/2.0/BrainStem/glTF/BrainStem.gltf but
	// does not work properly... I also tried with bit-shifting src[0...3] or
	// src[3...0] into dst, to no avail. HB
	template<>
	LL_INLINE void copyVec4<U16, U64>(U16* src, U64& dst)
	{
		U64* p = (U64*)src;
		dst = *p;
	}

	template<>
	LL_INLINE void copyMat4<F32, glm::mat4>(F32* src, glm::mat4& dst)
	{
		dst = glm::make_mat4(src);
	}

	//=========================================================================

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyScalar(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyScalar(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyVec2(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyVec2(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyVec3(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyVec3(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyVec4(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyVec4(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyMat2(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyMat2(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyMat3(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyMat3(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	// Copy from src to dst, stride is the number of bytes between each element
	// in src, count is number of elements to copy.
	template<class S, class T>
	LL_INLINE void copyMat4(S* src, LLStrider<T> dst, S32 stride, S32 count)
	{
		for (S32 i = 0; i < count; ++i)
		{
			copyMat4(src, *dst++);
			src = (S*)((U8*)src + stride);
		}
	}

	LL_NO_INLINE void unsupported_accessor_type(const char* func_sig, U8 type);

	template<class S, class T>
	LL_INLINE void copy(Asset& asset, Accessor& accessor, const S* src,
						LLStrider<T>& dst, S32 bstride)
	{
		if (accessor.mType == Accessor::Type::SCALAR)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 1 : bstride;
			copyScalar((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::VEC2)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 2 : bstride;
			copyVec2((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::VEC3)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 3 : bstride;
			copyVec3((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::VEC4)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 4 : bstride;
			copyVec4((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::MAT2)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 4 : bstride;
			copyMat2((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::MAT3)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 9 : bstride;
			copyMat3((S*)src, dst, stride, accessor.mCount);
		}
		else if (accessor.mType == Accessor::Type::MAT4)
		{
			S32 stride = bstride == 0 ? sizeof(S) * 16 : bstride;
			copyMat4((S*)src, dst, stride, accessor.mCount);
		}
		else
		{
			unsupported_accessor_type(LL_FUNC, (U8)accessor.mType);
		}
	}

	LL_NO_INLINE void invalid_component_type(const char* func_sig, U8 type);

	// Copy data from accessor to strider
	template<class T>
	LL_INLINE void copy(Asset& asset, Accessor& accessor, LLStrider<T>& dst)
	{
		const BufferView& buffview = asset.mBufferViews[accessor.mBufferView];
		const Buffer& buffer = asset.mBuffers[buffview.mBuffer];
		const U8* srcp = buffer.mData.data() + buffview.mByteOffset +
						 accessor.mByteOffset;
		const S32 stride = buffview.mByteStride;
		switch ((Accessor::ComponentType)accessor.mComponentType)
		{
			case Accessor::ComponentType::BYTE:
				copy(asset, accessor, (const S8*)srcp, dst, stride);
				break;

			case Accessor::ComponentType::UNSIGNED_BYTE:
				copy(asset, accessor, (const U8*)srcp, dst, stride);
				break;

			case Accessor::ComponentType::SHORT:
				copy(asset, accessor, (const S16*)srcp, dst, stride);
				break;

			case Accessor::ComponentType::UNSIGNED_SHORT:
				copy(asset, accessor, (const U16*)srcp, dst, stride);
				break;

			case Accessor::ComponentType::UNSIGNED_INT:
				copy(asset, accessor, (const U32*)srcp, dst, stride);
				break;

			case Accessor::ComponentType::FLOAT:
				copy(asset, accessor, (const F32*)srcp, dst, stride);
				break;

			default:
				invalid_component_type(LL_FUNC, (U32)accessor.mComponentType);
		}
	}

	// Copy data from accessor to vector
	template<class T>
	LL_INLINE void copy(Asset& asset, Accessor& accessor, std::vector<T>& dst)
	{
		dst.resize(accessor.mCount);
		LLStrider<T> strider = dst.data();
		copy(asset, accessor, strider);
	}

	//=========================================================================
	// boost::json copying utilities

	// To/from Value
	template<typename T>
	LL_INLINE bool copy(const Value& src, T& dst)
	{
		dst = src;
		return true;
	}

	template<typename T>
	LL_INLINE bool write(const T& src, Value& dst)
	{
		dst = Object();
		src.serialize(dst.as_object());
		return true;
	}

	template<typename T>
	LL_INLINE bool copy(const Value& src, fast_hmap<std::string, T>& dst)
	{
		if (!src.is_object())
		{
			return false;
		}
		const Object& obj = src.as_object();
		for (const auto& [key, value] : obj)
		{
			copy<T>(value, dst[key]);
		}
		return true;
	}

	template<typename T>
	LL_INLINE bool write(const fast_hmap<std::string, T>& src, Value& dst)
	{
		Object obj;
		for (const auto& [key, value] : src)
		{
			Value v;
			if (!write<T>(value, v))
			{
				return false;
			}
			obj[key] = v;
		}
		dst = obj;
		return true;
	}

	// To/from array
	template<typename T>
	LL_INLINE bool copy(const Value& src, std::vector<T>& dst)
	{
		if (!src.is_array())
		{
			return false;
		}
		const boost::json::array& arr = src.get_array();
		dst.resize(arr.size());
		for (size_t i = 0; i < arr.size(); ++i)
		{
			copy(arr[i], dst[i]);
		}
		return true;
	}

	template<typename T>
	LL_INLINE bool write(const std::vector<T>& src, Value& dst)
	{
		boost::json::array arr;
		for (const T& t : src)
		{
			Value v;
			if (!write(t, v))
			{
				return false;
			}
			arr.push_back(v);
		}
		dst = arr;
		return true;
	}

	// To/from object member
	template<typename T>
	LL_INLINE bool copy(const Object& src, const char* member, T& dst)
	{
		auto it = src.find(member);
		return it != src.end() && copy(it->value(), dst);
	}

	// Always write a member to an object without checking default
	template<typename T>
	LL_INLINE bool write_always(const T& src, const char* member, Object& dst)
	{
		Value& v = dst[member];
		if (!write(src, v))
		{
			dst.erase(member);
			return false;
		}
		return true;
	}

	// For internal use only: use copy_extensions instead.
	template<typename T>
	LL_INLINE bool _copy_extension(const Object& exts, const char* member,
								   T* dst)
	{
		return exts.contains(member) && copy(exts.at(member), *dst);
	}

	// Copies all extensions from src.extensions to provided destinations
	// Usage:
	//	copy_extensions(src,"KHR_materials_unlit", &mUnlit,
	//					"KHR_materials_pbrSpecularGlossiness",
	//					&mPbrSpecularGlossiness);
	// Returns true if any of the extensions are copied.
	template<class... Types>
	LL_INLINE bool copy_extensions(const Value& src, Types... args)
	{
		if (!src.is_object())
		{
			return false;
		}
		// Extract the extensions object (do not assume it exists and verify
		// that it is an object)
		Object obj = src.get_object();
		if (!obj.contains("extensions"))
		{
			return false;
		}
		const Value& exts = obj.at("extensions");
		if (!exts.is_object())
		{
			return false;
		}
		const Object& ext_obj = exts.as_object();
		bool success = false;
		size_t count = sizeof...(args);
		for (size_t i = 0; i < count; i += 2)
		{
			if (_copy_extension(ext_obj, args...))
			{
				success = true;
			}
		}
		return success;
	}

	// For internal use only: use write_extensions instead.
	template<typename T>
	LL_INLINE bool _write_extension(Object& exts, T* src, const char* member)
	{
		if (src->mPresent)
		{
			Value v;
			if (write(*src, v))
			{
				exts[member] = v;
				return true;
			}
		}
		return false;
	}

	// Writes all extensions to dst.extensions
	// Usage:
	//	write_extensions(dst, mUnlit, "KHR_materials_unlit",
	//					 mPbrSpecularGlossiness,
	//					 "KHR_materials_pbrSpecularGlossiness");
	// Returns true if any of the extensions are written.
	template<class... Types>
	LL_INLINE bool write_extensions(Object& dst, Types... args)
	{
		bool success = false;
		Object exts;
		size_t count = sizeof...(args) - 1;
		for (size_t i = 0; i < count; i += 2)
		{
			if (_write_extension(exts, args...))
			{
				success = true;
			}
		}
		if (success)
		{
			dst["extensions"] = exts;
		}
		return success;
	}

	// Conditionally write a member to an object if the member is not the
	// default value
	template<typename T>
	LL_INLINE bool write(const T& src, const char* member, Object& dst,
						 const T& default_value = T())
	{
		return src != default_value && write_always(src, member, dst);
	}

	template<typename T>
	LL_INLINE bool write(const fast_hmap<std::string, T>& src,
						 const char* member, Object& dst,
						 const fast_hmap<std::string, T>& default_value =
							fast_hmap<std::string, T>())
	{
		if (src.empty())
		{
			return false;
		}
		Value v;
		if (!write<T>(src, v))
		{
			return false;
		}
		dst[member] = v;
		return true;
	}

	template<typename T>
	LL_INLINE bool write(const std::vector<T>& src, const char* member,
						 Object& dst,
						 const std::vector<T>& default_value =
							std::vector<T>())
	{
		if (src.empty())
		{
			return false;
		}
		Value v;
		if (!write(src, v))
		{
			return false;
		}
		dst[member] = v;
		return true;
	}

	template<typename T>
	LL_INLINE bool copy(const Value& src, const char* member, T& dst)
	{
		if (!src.is_object())
		{
			return false;
		}
		const Object& obj = src.as_object();
		return copy(obj, member, dst);
	}

	// Accessor::ComponentType
	template<>
	LL_INLINE bool copy(const Value& src, Accessor::ComponentType& dst)
	{
		if (!src.is_int64())
		{
			return false;
		}
		dst = (Accessor::ComponentType)src.get_int64();
		return true;
	}

	template<>
	LL_INLINE bool write(const Accessor::ComponentType& src, Value& dst)
	{
		dst = (S32)src;
		return true;
	}

	// Primitive::Mode
	template<>
	LL_INLINE bool copy(const Value& src, Primitive::Mode& dst)
	{
		if (!src.is_int64())
		{
			return false;
		}
		dst = (Primitive::Mode)src.get_int64();
		return true;
	}

	template<>
	LL_INLINE bool write(const Primitive::Mode& src, Value& dst)
	{
		dst = (S32)src;
		return true;
	}

	LL_NO_INLINE void json_error(const char* func_sig,
								 const std::error_code& ec);

	// GLM types
	template<>
	LL_INLINE bool copy(const Value& src, glm::vec4& dst)
	{
		if (src.is_array())
		{
			const boost::json::array& arr = src.as_array();
			if (arr.size() == 4)
			{
				glm::vec4 t;
				std::error_code ec;
				t.x = arr[0].to_number<F32>(ec);
				if (!ec)
				{
					t.y = arr[1].to_number<F32>(ec);
					if (!ec)
					{
						t.z = arr[2].to_number<F32>(ec);
						if (!ec)
						{
							t.w = arr[3].to_number<F32>(ec);
						}
					}
				}
				if (ec)
				{
					json_error(LL_FUNC, ec);
					return false;
				}
				dst = t;
				return true;
			}
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const glm::vec4& src, Value& dst)
	{
		dst = boost::json::array();
		boost::json::array& arr = dst.get_array();
		arr.resize(4);
		arr[0] = src.x;
		arr[1] = src.y;
		arr[2] = src.z;
		arr[3] = src.w;
		return true;
	}

	template<>
	LL_INLINE bool copy(const Value& src, glm::quat& dst)
	{
		if (src.is_array())
		{
			const boost::json::array& arr = src.as_array();
			if (arr.size() == 4)
			{
				glm::quat t;
				std::error_code ec;
				t.x = arr[0].to_number<F32>(ec);
				if (!ec)
				{
					t.y = arr[1].to_number<F32>(ec);
					if (!ec)
					{
						t.z = arr[2].to_number<F32>(ec);
						if (!ec)
						{
							t.w = arr[3].to_number<F32>(ec);
						}
					}
				}
				if (ec)
				{
					json_error(LL_FUNC, ec);
					return false;
				}
				dst = t;
				return true;
			}
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const glm::quat& src, Value& dst)
	{
		dst = boost::json::array();
		boost::json::array& arr = dst.get_array();
		arr.resize(4);
		arr[0] = src.x;
		arr[1] = src.y;
		arr[2] = src.z;
		arr[3] = src.w;
		return true;
	}

	template<>
	LL_INLINE bool copy(const Value& src, glm::vec2& dst)
	{
		if (src.is_array())
		{
			const boost::json::array& arr = src.as_array();
			if (arr.size() == 2)
			{
				glm::vec2 t;
				std::error_code ec;
				t.x = arr[0].to_number<F32>(ec);
				if (!ec)
				{
					t.y = arr[1].to_number<F32>(ec);
				}
				if (ec)
				{
					json_error(LL_FUNC, ec);
					return false;
				}
				dst = t;
				return true;
			}
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const glm::vec2& src, Value& dst)
	{
		dst = boost::json::array();
		boost::json::array& arr = dst.as_array();
		arr.resize(2);
		arr[0] = src.x;
		arr[1] = src.y;
		return true;
	}

	template<>
	LL_INLINE bool write(const glm::vec3& src, Value& dst)
	{
		dst = boost::json::array();
		boost::json::array& arr = dst.as_array();
		arr.resize(3);
		arr[0] = src.x;
		arr[1] = src.y;
		arr[2] = src.z;
		return true;
	}

	template<>
	LL_INLINE bool copy(const Value& src, glm::vec3& dst)
	{
		if (src.is_array())
		{
			const boost::json::array& arr = src.as_array();
			if (arr.size() == 3)
			{
				glm::vec3 t;
				std::error_code ec;
				t.x = arr[0].to_number<F32>(ec);
				if (!ec)
				{
					t.y = arr[1].to_number<F32>(ec);
					if (!ec)
					{
						t.z = arr[2].to_number<F32>(ec);
					}
				}
				if (ec)
				{
					json_error(LL_FUNC, ec);
					return false;
				}
				dst = t;
				return true;
			}
		}
		return false;
	}

	template<>
	LL_INLINE bool copy(const Value& src, glm::mat4& dst)
	{
		if (src.is_array())
		{
			const boost::json::array& arr = src.as_array();
			if (arr.size() == 16)
			{
				std::error_code ec;
				glm::mat4 t;
				F32* p = glm::value_ptr(t);
				for (U32 i = 0; i < 16; ++i)
				{
					p[i] = arr[i].to_number<F32>(ec);
					if (ec)
					{
						json_error(LL_FUNC, ec);
						return false;
					}
				}
				dst = t;
				return true;
			}
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const glm::mat4& src, Value& dst)
	{
		const F32* p = glm::value_ptr(src);
		dst = boost::json::array();
		boost::json::array& arr = dst.get_array();
		arr.resize(16);
		for (U32 i = 0; i < 16; ++i)
		{
			arr[i] = p[i];
		}
		return true;
	}

	// bool
	template<>
	LL_INLINE bool copy(const Value& src, bool& dst)
	{
		if (src.is_bool())
		{
			dst = src.get_bool();
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const bool& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// S32
	template<>
	LL_INLINE bool copy(const Value& src, S32& dst)
	{
		if (src.is_int64())
		{
			dst = src.get_int64();
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const S32& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// U32
	template<>
	LL_INLINE bool copy(const Value& src, U32& dst)
	{
		if (src.is_int64())
		{
			dst = src.get_int64();
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const U32& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// F32
	template<>
	LL_INLINE bool copy(const Value& src, F32& dst)
	{
		std::error_code ec;
		F32 t = src.to_number<F32>(ec);
		if (ec)
		{
			json_error(LL_FUNC, ec);
			return false;
		}
		dst = t;
		return true;
	}

	template<>
	LL_INLINE bool write(const F32& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// F64
	template<>
	LL_INLINE bool copy(const Value& src, F64& dst)
	{
		std::error_code ec;
		F64 t = src.to_number<F64>(ec);
		if (ec)
		{
			json_error(LL_FUNC, ec);
			return false;
		}
		dst = t;
		return true;
	}

	template<>
	LL_INLINE bool write(const F64& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// std::string
	template<>
	LL_INLINE bool copy(const Value& src, std::string& dst)
	{
		if (src.is_string())
		{
			dst = src.get_string().c_str();
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const std::string& src, Value& dst)
	{
		dst = src;
		return true;
	}

	// Accessor::Type
	template<>
	LL_INLINE bool copy(const Value& src, Accessor::Type& dst)
	{
		if (src.is_string())
		{
			dst = gltf_type_to_enum(src.get_string().c_str());
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const Accessor::Type& src, Value& dst)
	{
		dst = enum_to_gltf_type(src);
		return true;
	}

	// Material::AlphaMode
	template<>
	LL_INLINE bool copy(const Value& src,Material::AlphaMode& dst)
	{
		if (src.is_string())
		{
			dst = gltf_alpha_mode_to_enum(src.get_string().c_str());
			return true;
		}
		return false;
	}

	template<>
	LL_INLINE bool write(const Material::AlphaMode& src, Value& dst)
	{
		dst = enum_to_gltf_alpha_mode(src);
		return true;
	}
}

#if LL_CLANG
# pragma clang diagnostic pop
#endif

#endif	// LL_LLGLTFBUFFERUTIL_H
