/**
 * @file llpacketring.cpp
 * @brief implementation of LLPacketRing class for a packet.
 *
 * $LicenseInfo:firstyear=2001&license=viewergpl$
 *
 * Copyright (c) 2001-2009, Linden Research, Inc.
 *
 * Second Life Viewer Source Code
 * The source code in this file ("Source Code") is provided by Linden Lab
 * to you under the terms of the GNU General Public License, version 2.0
 * ("GPL"), unless you have obtained a separate licensing agreement
 * ("Other License"), formally executed by you and Linden Lab.  Terms of
 * the GPL can be found in doc/GPL-license.txt in this distribution, or
 * online at http://secondlifegrid.net/programs/open_source/licensing/gplv2
 *
 * There are special exceptions to the terms and conditions of the GPL as
 * it is applied to this Source Code. View the full text of the exception
 * in the file doc/FLOSS-exception.txt in this software distribution, or
 * online at
 * http://secondlifegrid.net/programs/open_source/licensing/flossexception
 *
 * By copying, modifying or distributing this software, you acknowledge
 * that you have read and understood your obligations described above,
 * and agree to abide by those obligations.
 *
 * ALL LINDEN LAB SOURCE CODE IS PROVIDED "AS IS." LINDEN LAB MAKES NO
 * WARRANTIES, EXPRESS, IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY,
 * COMPLETENESS OR PERFORMANCE.
 * $/LicenseInfo$
 */

#include "linden_common.h"

#include "llpacketring.h"

#if LL_WINDOWS
# include <winsock2.h>
#else
# include <sys/socket.h>
# include <netinet/in.h>
#endif

#include "llmessage.h"
#include "llnet.h"
#include "llproxy.h"
#include "llrand.h"
#include "lltimer.h"

///////////////////////////////////////////////////////////////////////////////
// LLPacketBuffer class. Used to be in its own llpacketbuffer.h/cpp module, but
// is only used by LLPacketRing, so I moved it here. HB
///////////////////////////////////////////////////////////////////////////////

class LLPacketBuffer
{
public:
	LLPacketBuffer(const LLHost& host, const char* datap, S32 size);
	LL_INLINE LLPacketBuffer(S32 socket)			{ init(socket); }

	void init(S32 socket);
	void init(const char* bufferp, S32 data_size, const LLHost& host);

	LL_INLINE S32 getSize() const					{ return mSize; }
	LL_INLINE const char* getData() const			{ return mData; }
	LL_INLINE LLHost getHost() const				{ return mHost; }
	LL_INLINE LLHost getReceivingInterface() const	{ return mReceivingIF; }

protected:
	LLHost	mHost;					// Source/dest IP and port
	LLHost	mReceivingIF;			// Source/dest IP and port
	char	mData[NET_BUFFER_SIZE];	// Packet data
	S32		mSize;					// Size of buffer in bytes
};

LLPacketBuffer::LLPacketBuffer(const LLHost& host, const char* datap, S32 size)
:	mHost(host)
{
	mSize = 0;
	mData[0] = '!';

	if (size > NET_BUFFER_SIZE)
	{
		llerrs << "Sending packet > " << NET_BUFFER_SIZE << " of size " << size
			   << llendl;
	}
	if (datap)
	{
		memcpy(mData, datap, size);
		mSize = size;
	}
}

void LLPacketBuffer::init(S32 socket)
{
	mSize = receive_packet(socket, mData);
	mHost = get_sender();
	mReceivingIF = get_receiving_interface();
}

void LLPacketBuffer::init(const char* bufferp, S32 size, const LLHost& host)
{
	if (size > NET_BUFFER_SIZE)
	{
		llerrs << "Cannot initialize packet with more than " << NET_BUFFER_SIZE
			   << " bytes: requested size was: " << size << llendl;
	}
	memcpy(mData, bufferp, size);
	mSize = size;
	mHost = host;
	mReceivingIF = get_receiving_interface();
}

///////////////////////////////////////////////////////////////////////////////
// LLPacketRing class
///////////////////////////////////////////////////////////////////////////////

constexpr U32 MAX_BUFFER_RING_SIZE = 1024;
constexpr U32 DEFAULT_BUFFER_RING_SIZE = 256;

LLPacketRing::LLPacketRing()
:	mHeadIndex(0),
	mNumBufferedPackets(0),
	mNumBufferedBytes(0),
	mActualBytesIn(0),
	mActualBytesOut(0)
{
	LLHost invalid_host;
	mPacketRing.reserve(DEFAULT_BUFFER_RING_SIZE);
	for (U32 i = 0; i < DEFAULT_BUFFER_RING_SIZE; ++i)
	{
		mPacketRing.emplace_back(new LLPacketBuffer(invalid_host, NULL, 0));
	}
}

LLPacketRing::~LLPacketRing()
{
	for (size_t i = 0, count = mPacketRing.size(); i < count; ++i)
	{
		delete mPacketRing[i];
	}
}

bool LLPacketRing::sendPacket(S32 socket, const char* datap, S32 data_size,
							  LLHost host)
{
	mActualBytesOut += data_size;

	if (!LLProxy::isSOCKSProxyEnabled())
	{
		return send_packet(socket, datap, data_size, host.getAddress(),
						   host.getPort());
	}

	char headered_send_buffer[NET_BUFFER_SIZE + SOCKS_HEADER_SIZE];

	proxywrap_t* headerp = (proxywrap_t*)((void*)&headered_send_buffer);
	headerp->rsv = 0;
	headerp->addr = host.getAddress();
	headerp->port = htons(host.getPort());
	headerp->atype = ADDRESS_IPV4;
	headerp->frag = 0;

	memcpy(headered_send_buffer + SOCKS_HEADER_SIZE, datap, data_size);

	LLHost proxyhost = LLProxy::getInstance()->getUDPProxy();
	return send_packet(socket, headered_send_buffer,
					   data_size + SOCKS_HEADER_SIZE,
					   proxyhost.getAddress(), proxyhost.getPort());
}

U32 LLPacketRing::receivePacket(S32 socket, char* datap)
{
	return mNumBufferedPackets > 0 ? receiveBufferedPacket(datap)
								   : receiveNetworkPacket(socket, datap);
}

S32 LLPacketRing::receiveNetworkPacket(S32 socket, char* datap)
{
	S32 packet_size;

	if (LLProxy::isSOCKSProxyEnabled())
	{
		char buffer[NET_BUFFER_SIZE + SOCKS_HEADER_SIZE];
		packet_size = receive_packet(socket, buffer);
		if (packet_size > 0)
		{
			mActualBytesIn += packet_size;
		}
		if (packet_size > SOCKS_HEADER_SIZE)
		{
			// *FIX: we are assuming ATYP is 0x01 (IPv4), not 0x03 (hostname)
			// or 0x04 (IPv6)
			packet_size -= SOCKS_HEADER_SIZE; // The unwrapped packet size
			memcpy(datap, buffer + SOCKS_HEADER_SIZE, packet_size);
			proxywrap_t* headerp = (proxywrap_t*)((void*)buffer);
			mLastSender.setAddress(headerp->addr);
			mLastSender.setPort(ntohs(headerp->port));
			mLastReceivingIF = get_receiving_interface();
		}
		else
		{
			packet_size = 0;
		}
	}
	else
	{
		packet_size = receive_packet(socket, datap);
		if (packet_size > 0)
		{
			mActualBytesIn += packet_size;
			mLastSender = get_sender();
			mLastReceivingIF = get_receiving_interface();
		}
	}

	return packet_size;
}

S32 LLPacketRing::receiveBufferedPacket(char* datap)
{
	U32 ring_size = mPacketRing.size();
	U32 packet_index = (mHeadIndex + ring_size - mNumBufferedPackets) %
					   ring_size;
	LLPacketBuffer* packetp = mPacketRing[packet_index];
	S32 packet_size = packetp->getSize();
	mLastSender = packetp->getHost();
	mLastReceivingIF = packetp->getReceivingInterface();

	--mNumBufferedPackets;
	mNumBufferedBytes -= packet_size;
	if (!mNumBufferedPackets && mNumBufferedBytes)
	{
		llwarns << "No buffered packet left but number of buffered bytes in non zero: "
				<< mNumBufferedBytes << llendl;
	}

	if (packet_size > 0 && packetp->getData())
	{
		memcpy(datap, packetp->getData(), packet_size);
	}
	else if (packet_size > 0)
	{
		llwarns << "Received a packet with non-zero size and no data. Zeroing data."
				<< llendl;
		memset(datap, 0, packet_size);
	}

	return packet_size;
}

S32 LLPacketRing::bufferInboundPacket(S32 socket)
{
	if (mNumBufferedPackets == (U32)mPacketRing.size() &&
		mNumBufferedPackets < MAX_BUFFER_RING_SIZE)
	{
		expandRing();
	}

	LLPacketBuffer* packetp = mPacketRing[mHeadIndex];
	S32 old_packet_size = packetp->getSize();

	S32 packet_size;

	if (LLProxy::isSOCKSProxyEnabled())
	{
		char buffer[NET_BUFFER_SIZE + SOCKS_HEADER_SIZE];
		packet_size = receive_packet(socket, buffer);
		if (packet_size > 0)
		{
			mActualBytesIn += packet_size;
			if (packet_size > SOCKS_HEADER_SIZE)
			{
				packet_size -= SOCKS_HEADER_SIZE; // The unwrapped packet size
				// *FIX: we are assuming ATYP is 0x01 (IPv4), not 0x03
				// (hostname) or 0x04 (IPv6)
				proxywrap_t* headerp = (proxywrap_t*)((void*)buffer);
				LLHost sender;			
				sender.setAddress(headerp->addr);
				sender.setPort(ntohs(headerp->port));
				packetp->init(buffer + SOCKS_HEADER_SIZE, packet_size, sender);

				mHeadIndex = (mHeadIndex + 1) % U32(mPacketRing.size());
				if (mNumBufferedPackets < MAX_BUFFER_RING_SIZE)
				{
					++mNumBufferedPackets;
					mNumBufferedBytes += packet_size;
				}
				else
				{
					// We overwrote an older packet
					mNumBufferedBytes += packet_size - old_packet_size;
				}
			}
			else
			{
				packet_size = 0;
			}
		}
	}
	else
	{
		packetp->init(socket);
		packet_size = packetp->getSize();
		if (packet_size > 0)
		{
			mHeadIndex = (mHeadIndex + 1) % U32(mPacketRing.size());
			mActualBytesIn += packet_size;
			if (mNumBufferedPackets < MAX_BUFFER_RING_SIZE)
			{
				++mNumBufferedPackets;
				mNumBufferedBytes += packet_size;
			}
			else
			{
				// We overwrote an older packet
				mNumBufferedBytes += packet_size - old_packet_size;
			}
		}
	}

	return packet_size;
}

// Drains to buffer
U32 LLPacketRing::drainSocket(LLMessageSystem* msgsys, S32 socket)
{
	S32 packet_size = 1;
	S32 num_loops = 0;
	S32 old_num_packets = mNumBufferedPackets;
	while (packet_size > 0)
	{
		packet_size = bufferInboundPacket(socket);
		++num_loops;
	}
	S32 dropped = (num_loops - 1 + old_num_packets) - mNumBufferedPackets;
	if (dropped > 0)
	{
		msgsys->mDroppedPackets += dropped;
		llwarns << "Dropped " << dropped << " UDP packets." << llendl;
	}
	return mNumBufferedPackets;
}

bool LLPacketRing::expandRing()
{
	constexpr U32 BUFFER_RING_EXPANSION = 256;
	U32 old_size = mPacketRing.size();
	U32 new_size = llmin(old_size + BUFFER_RING_EXPANSION,
						 MAX_BUFFER_RING_SIZE);
	if (new_size == old_size)
	{
		return false;	// mPacketRing is already maxed out
	}

	// Make a larger ring and copy packet pointers
	packet_vec_t new_ring;
	new_ring.reserve(new_size);
	for (U32 i = 0; i < old_size; ++i)
	{
		new_ring.emplace_back(mPacketRing[(mHeadIndex + i) % old_size]);
	}
	// Allocate new packets for the remainder of new_ring
	LLHost invalid_host;
	for (U32 i = old_size; i < new_size; ++i)
	{
		new_ring.emplace_back(new LLPacketBuffer(invalid_host, NULL, 0));
	}
	// Swap the rings and reset mHeadIndex
	mPacketRing.swap(new_ring);
	mHeadIndex = mNumBufferedPackets;

	return true;
}

F32 LLPacketRing::getBufferLoadRate() const
{
	// Goes up to MAX_BUFFER_RING_SIZE
	return (F32)mNumBufferedPackets / (F32)DEFAULT_BUFFER_RING_SIZE;
}
