/**
 * @file llpacketring.h
 * @brief definition of LLPacketRing class for implementing a resend,
 * drop, or delay in packet transmissions
 *
 * $LicenseInfo:firstyear=2001&license=viewergpl$
 *
 * Copyright (c) 2001-2009, Linden Research, Inc.
 *
 * Second Life Viewer Source Code
 * The source code in this file ("Source Code") is provided by Linden Lab
 * to you under the terms of the GNU General Public License, version 2.0
 * ("GPL"), unless you have obtained a separate licensing agreement
 * ("Other License"), formally executed by you and Linden Lab.  Terms of
 * the GPL can be found in doc/GPL-license.txt in this distribution, or
 * online at http://secondlifegrid.net/programs/open_source/licensing/gplv2
 *
 * There are special exceptions to the terms and conditions of the GPL as
 * it is applied to this Source Code. View the full text of the exception
 * in the file doc/FLOSS-exception.txt in this software distribution, or
 * online at
 * http://secondlifegrid.net/programs/open_source/licensing/flossexception
 *
 * By copying, modifying or distributing this software, you acknowledge
 * that you have read and understood your obligations described above,
 * and agree to abide by those obligations.
 *
 * ALL LINDEN LAB SOURCE CODE IS PROVIDED "AS IS." LINDEN LAB MAKES NO
 * WARRANTIES, EXPRESS, IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY,
 * COMPLETENESS OR PERFORMANCE.
 * $/LicenseInfo$
 */

#ifndef LL_LLPACKETRING_H
#define LL_LLPACKETRING_H

#include <queue>

#include "llhost.h"

class LLMessageSystem;
class LLPacketBuffer;

class LLPacketRing
{
protected:
	LOG_CLASS(LLPacketRing);

public:
	LLPacketRing();
    ~LLPacketRing();

	// Receives one packet: either buffered or from the socket.
	U32 receivePacket(S32 socket, char* datap);

	// Sends one packet.
	bool sendPacket(S32 h_socket, const char* datap, S32 data_size,
					LLHost host);

	// Drains packets from socket and returns final mNumBufferedPackets.
	U32 drainSocket(LLMessageSystem* msgsys, S32 socket);

	LL_INLINE LLHost getLastSender() const			{ return mLastSender; }

	LL_INLINE LLHost getLastReceivingInterface() const
	{
		return mLastReceivingIF;
	}

	LL_INLINE U32 getActualInBytes() const			{ return mActualBytesIn; }
	LL_INLINE U32 getActualOutBytes() const			{ return mActualBytesOut; }

	LL_INLINE U32 getAndResetActualInBits()
	{
		U32 bits = 8 * mActualBytesIn;
		mActualBytesIn = 0;
		return bits;
	}

	LL_INLINE U32 getAndResetActualOutBits()
	{
		U32 bits = 8 * mActualBytesOut;
		mActualBytesOut = 0;
		return bits;
	}

	LL_INLINE U32 getNumBufferedPackets() const
	{
		return mNumBufferedPackets;
	}

	LL_INLINE U32 getNumBufferedBytes() const
	{
		return mNumBufferedBytes;
	}

	// Returns 0 (empty) to 4 (1 - default size is full)
	F32 getBufferLoadRate() const;

private:
	// These methods return of received packet, zero or less if no packet found
	S32 receiveNetworkPacket(S32 socket, char* datap);
	S32 receiveBufferedPacket(char* datap);

	// Returns packet_size of packet buffered.
	S32 bufferInboundPacket(S32 socket);

	// Returns 'true' if ring was expanded
	bool expandRing();

protected:
	typedef std::vector<LLPacketBuffer*> packet_vec_t;
	packet_vec_t	mPacketRing;

	LLHost			mLastSender;
	LLHost			mLastReceivingIF;

	U32				mHeadIndex;
	U32				mNumBufferedPackets;
	U32				mNumBufferedBytes;
	U32				mActualBytesIn;
	U32				mActualBytesOut;
};

#endif
