/**
 * @file windows_volume_catcher.cpp
 * @brief A Windows implementation of volume level control of all audio
 * channels opened by a process.
 *
 * @cond
 * $LicenseInfo:firstyear=2010&license=viewerlgpl$
 * Second Life Viewer Source Code
 * Copyright (C) 2010, Linden Research, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License only.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Linden Research, Inc., 945 Battery Street, San Francisco, CA  94111  USA
 * $/LicenseInfo$
 * @endcond
 */

#include "linden_common.h"

#include <mmeapi.h>

#include "volume_catcher.h"

#include "llsingleton.h"

class VolumeCatcherImpl : public LLSingleton<VolumeCatcherImpl>
{
	friend class LLSingleton<VolumeCatcherImpl>;

public:
	void setVolume(F32 volume);

	// Remember pan for calculating individual channel levels later
	LL_INLINE void setPan(F32 pan)		{ mPan = pan; }

private:
	// This is a singleton class: both callers and the component implementation
	// should use getInstance() to find the instance.
	VolumeCatcherImpl();
	~VolumeCatcherImpl() override = default;

private:
	F32 				mVolume;
	F32 				mPan;
};

VolumeCatcherImpl::VolumeCatcherImpl()
:	mVolume(1.f),	// default volume is max
	mPan(0.f)		// default pan is centered
{
}

void VolumeCatcherImpl::setVolume(F32 volume)
{
	mVolume = volume;

	// Set both left/right to same volume.
	// *TODO: use pan value to set independently
	DWORD left_channel = (DWORD)(mVolume * 65535.0f);
	DWORD right_channel = (DWORD)(mVolume * 65535.0f);
	DWORD hw_volume = left_channel << 16 | right_channel;
	::waveOutSetVolume(NULL, hw_volume);
}

/////////////////////////////////////////////////////

VolumeCatcher::VolumeCatcher()
{
	pimpl = VolumeCatcherImpl::getInstance();
}

void VolumeCatcher::setVolume(F32 volume)
{
	pimpl->setVolume(volume);
}

void VolumeCatcher::setPan(F32 pan)
{
	pimpl->setPan(pan);
}

void VolumeCatcher::pump()
{
	// No periodic tasks are necessary for this implementation.
}
